import matplotlib.pyplot as plt
import numpy             as np
import pandas            as pd
import seaborn           as sns
from matplotlib          import get_backend
from scipy.stats         import skewtest, kurtosistest

###############################################################################
#Globals
###############################################################################
#Refer for details: https://seaborn.pydata.org/tutorial/color_palettes.html
palette_types = {'color':     lambda n_colors, palette='bright', **kwargs : sns.color_palette(palette=palette, n_colors=n_colors),
                 'light':     lambda n_colors, color, **kwargs            : sns.light_palette(n_colors=n_colors+2, color=color, **kwargs)[2:],
                 'dark' :     lambda n_colors, color, **kwargs            : sns.dark_palette( n_colors=n_colors+2, color=color, **kwargs)[2:],
                 'cubehelix': lambda n_colors, **kwargs                   : sns.cubehelix_palette(n_colors=n_colors, **kwargs),
                 'diverging': lambda n, **kwargs                          : sns.diverging_palette(n=n, **kwargs),
                 }    
#Refer for details: https://xkcd.com/color/rgb/
all_colors    = sns.colors.xkcd_rgb

###############################################################################
#Import
###############################################################################
def import_trace(files, keys=[], **pd_args):
    '''Used for traces that have been saved to csv files.
    
    :param files: A list of file names.
    :param keys: A list of keys to index the traces.
    :param pd_args: Keyword arguments for pandas.read_csv.
    :return traces: Returns a dict of traces.
        
    '''
    data   = {}
    for i in range(len(files)):
        key       = keys[i+1] if keys else i+1
        data[key] = pd.read_csv(files[i],  **pd_args)
    return data    

###############################################################################
#Skewness and Kurtosis
###############################################################################
def check_skewness(traces, output='df'):
    '''Calculates skewness of distribution of sampled parameters
    
    :param traces: A dict of traces.
    :param output: Causes the return value to be formatted as DataFrame.
    :return result: A DataFrame if output is "df" and a dict otherwise.

    '''
    return scipy_test(skewtest, traces, output='df')

def check_kurtosis(traces, output='df'):
    '''Calculates kurtosis of distribution of sampled parameters
    
    :param traces: A dict of traces.
    :param output: Causes the return value to be formatted as DataFrame.
    :return result: A DataFrame if output is "df" and a dict otherwise.

    '''
    return scipy_test(kurtosistest, traces, output='df')

def scipy_test(test_func, traces, output='df'):
    '''
    :meta private:
    '''
    result    = {}
    for label in traces:
        trace       = traces[label]
        test_result = test_func(trace, axis=0, nan_policy='omit')
        
        if output == 'df':
            variables     = trace.columns.to_list()
            df            = pd.DataFrame( test_result, columns=variables, index=('stats', 'pval'))
            result[label] = df
        else:
            result[label] = test_result
    
    return result
        
###############################################################################
#Wrapper Functions for Bivariate Plots
###############################################################################
def pairplot_steps(traces, pairs, figs=[], AX={}, gradient=5, palette={}, legend_args={}, plot_args={'marker': '+', 'linewidth':0}, palette_type='light'):
    '''Generates a pair plot between two parameters.
    '''
    return pairplot_wrapper('plot', traces=traces, pairs=pairs, figs=figs, AX=AX, palette=palette, gradient=gradient, palette_type=palette_type, legend_args=legend_args, plot_args=plot_args)

def pairplot_kde(traces, pairs, figs=[], AX={}, palette={}, legend_args={}, plot_args={}):
    '''Generates a pair plot between two parameters in kde form.
    '''
    return pairplot_wrapper('kde', traces=traces, pairs=pairs, figs=figs, AX=AX, palette=palette, gradient=1, palette_type='light', legend_args=legend_args, plot_args=plot_args)

def pairplot_wrapper(plot_func, traces, pairs, figs=[], AX={}, palette={}, gradient=1, palette_type='light', legend_args={}, plot_args={}):
    '''
    :meta private:
    '''
    pairs1       = [tuple(pair) for pair in pairs]
    palette1     = make_palette(traces, pairs1, palette, gradient=gradient, palette_type=palette_type)
    figs1, AX1   = figs, AX
    
    for label in traces:
        figs1, AX1 = plot_helper(plot_func, label, traces[label], pairs1, figs=figs1, AX=AX1, palette=palette1[label], legend_args=legend_args, plot_args=plot_args)
    
    [fs(fig) for fig in figs1]
    
    return figs1, AX1
    
###############################################################################
#Wrapper Functions for Univariate Plots
###############################################################################    
def plot_steps(traces, skip=[], figs=None, AX=None, palette=None, legend_args={}, plot_args={'marker': '+', 'linewidth':0}):
    '''Generates trace plot for parameters.
    
    :param traces: A dict of traces.
    :param skip: A list of parameters to not plot.
    :param figs: A list of figures for containing the plots. Default is None.
    :param AX: A dict of param - Axes object pairs. Default is None.
    :param palette: A dict of colors. Default is None.
    :param legend_args: A dict of arguments for the legend. Default is None.
    :param plot_args: A dict of arguments for plotting such as marker.
    :return result: Figure and Axes objects.
    
    '''
    return singleplot_wrapper('plot', traces, skip=skip, figs=figs, AX=AX, palette=palette, legend_args=legend_args, plot_args=plot_args)
    
def plot_kde(traces, skip=[], figs=[], AX={}, palette={}, legend_args={}, plot_args={'linewidth': 3}):
    '''Generates kde plot for parameters.
    
    :param traces: A dict of traces.
    :param skip: A list of parameters to not plot.
    :param figs: A list of figures for containing the plots. Default is None.
    :param AX: A dict of param - Axes object pairs. Default is None.
    :param palette: A dict of colors. Default is None.
    :param legend_args: A dict of arguments for the legend. Default is None.
    :param plot_args: A dict of arguments for plotting such as marker.
    :return result: Figure and Axes objects.
    
    '''
    return singleplot_wrapper('kde', traces, skip=skip, figs=figs, AX=AX, palette=palette, legend_args=legend_args, plot_args=plot_args)
    
def plot_hist(traces, skip=[], figs=[], AX={}, palette={}, legend_args={}, plot_args={}):
    '''Generates histogram plot for parameters.
    
    :param traces: A dict of traces.
    :param skip: A list of parameters to not plot.
    :param figs: A list of figures for containing the plots. Default is None.
    :param AX: A dict of param - Axes object pairs. Default is None.
    :param palette: A dict of colors. Default is None.
    :param legend_args: A dict of arguments for the legend. Default is None.
    :param plot_args: A dict of arguments for plotting such as marker.
    :return result: Figure and Axes objects.
    
    '''
    return singleplot_wrapper('hist', traces, skip=skip, figs=figs, AX=AX, palette=palette, legend_args=legend_args, plot_args=plot_args)
    
def singleplot_wrapper(plot_func, traces, skip=[], figs=[], AX={}, palette={}, legend_args={}, plot_args={}):
    '''
    :meta private:
    '''
    figs        = [] if figs is None else figs
    AX          = {} if AX   is None else AX
    palette     = {} if palette is None else palette
    legend_args = {} if legend_args is None else legend_args
    plot_args   = {} if plot_args  is None else plot_args 
    
    figs1, AX1, variables, first_label = setup_singleplot(traces, skip, figs, AX)
    palette1                           = make_palette(traces, variables, palette)
    legend_args1                       = legend_args if AX else {'labels': []}
    
    for label in traces:
        figs1, AX1 = plot_helper(plot_func, label, traces[label], variables, figs=figs1, AX=AX1, palette=palette1[label], legend_args=legend_args1, plot_args=plot_args)
    
    [fs(fig) for fig in figs1]
    
    if not AX:
        AX1[next(iter(AX1))].legend(**legend_args)
    return figs1, AX1

###############################################################################
#Supporting Functions for Wrapping
###############################################################################
def setup_singleplot(traces, skip, figs, AX):
    '''
    :meta private:
    '''
    first_label = next(iter(traces))
    variables   = [variable for variable in traces[first_label].columns.to_list() if variable not in skip]
    n           = len(variables)
    
    figs1 = figs if AX else [plt.figure()]
    if AX:
        AX1   = AX  
    else:
        AX1 = {variables[i]: figs1[0].add_subplot(n//2 + n%2, 2, i+1) for i in range(len(variables))}
    
    return figs1, AX1, variables, first_label

def make_palette(traces, variables, palette, gradient=1, palette_type='light'):
    '''
    :meta private:
    '''
    global palette_types
    if palette:
        if type(palette[next(iter(palette))]) == dict:
            #Assume colors have been fully specified
            palette1 = palette
        else:
            #Assume colors for each trace has been specified
            palette1 = {}
            for label in palette:
                base_color = palette[label]
                colors     = palette_types[palette_type](gradient, base_color)
                palette1[label] = {variable: colors for variable in variables}
                
    else:
        base_colors = palette_types['color'](len(traces), 'muted')
    
        palette1 = {}
        labels   = list(traces.keys())
        for i in range(len(labels)):
            label           = labels[i] if type(labels[i]) != list else tuple(labels[i])
            base_color      = base_colors[i]
            colors          = palette_types[palette_type](gradient, base_color)
            palette1[label] = {variable: colors for variable in variables}
    
    return palette1

###############################################################################
#Main Plotting
###############################################################################
def plot_helper(plot_func, label, trace, variables, figs=[], AX={}, palette={}, legend_args={}, plot_args={}):
    '''
    :meta private:
    '''
    n     = len(variables)    
    figs1 = figs if AX else [plt.figure()                              for i in range(n)]
    AX1   = AX   if AX else {variables[i]: figs1[i].add_subplot(1,1,1) for i in range(n)}

    lines = []
    for i in range(len(variables)):
        
        if type(variables[i]) == str:
            vx, vy = None, variables[i]
        else:
            vx, vy = variables[i]
            
        ax = AX1[variables[i]]
        ax.set(xlabel=vx, ylabel=vy)
        ax.ticklabel_format(style='sci', scilimits=(-2,3))
        
        colors = palette.get(variables[i], [all_colors['baby blue']])
        x      = np.array_split(trace[vx], len(colors)) if vx else [[]]*len(colors)
        y      = np.array_split(trace[vy], len(colors))

        mapped_lines   = [axis_plot(ax, plot_func, x=x[i], y=y[i], color=colors[i],  label=label if i==len(colors)-1 else '_nolabel',**plot_args) for i in range(len(colors))]

        lines.append(mapped_lines[-1])
    
    if legend_args: 
        try:
            if len(legend_args.get('labels', [])):
                legend_args['handles'] = lines
            for key in AX1:
                AX1[key].legend(**legend_args)
        except:
            pass
    return figs1, AX1

def axis_plot(ax, func, x, y, *args, **kwargs):
    '''
    :meta private:
    '''
    if func == 'plot':      
        if len(x):
            ax.plot(x, y, *args, **kwargs)
        else:
            ax.plot(y, *args, **kwargs)
    elif func == 'hist':
        ax.hist(y, *args, **kwargs)
    elif func == 'kde':
        if len(x):
            sns.kdeplot(data=y, data2=x, *args, ax=ax, **kwargs)
        else:
            sns.kdeplot(y, *args, ax=ax, **kwargs)
    return

def fs(figure):
    '''
    :meta private:
    '''
    try:
        plt.figure(figure.number)
        backend   = get_backend()
        manager   = plt.get_current_fig_manager()
        
        if backend == 'TkAgg':
            manager.resize(*manager.window.maxsize())
        
        elif backend == 'Qt5Agg' or backend == 'Qt4Agg': 
            manager.window.showMaximized()
        
        else:
            manager.frame.Maximize(True)
        plt.pause(0.03)
    except:
        pass
    return figure


from os.path import dirname, join

###############################################################################
#Globals
###############################################################################
model_functions_directory = join(dirname(__file__), 'model_functions')

###############################################################################
#Functions
###############################################################################
def model_to_code(core_model, use_numba=True, local=False):
    
    header = 'import numpy as np\nfrom numpy import log   as ln\nfrom numpy import log10 as log\nfrom numpy import exp\n'
    if use_numba:
        header += 'from numba import jit\n\n' + '@jit(nopython=True)\n'
    
    result  = header + 'def model_' + core_model['system_type'].replace(', ', '_') + '(y, t, params):\n' 
    
    states     = states_to_code(core_model['states'])
    params     = params_to_code(core_model['parameters'] + core_model['inputs'])
    equations  = equations_to_code(core_model['equations'], core_model['states'])
    result    += '\n\n'.join([states, params, equations])
    
    filename = core_model['system_type'].replace(', ', '_') + '.py'
    export_code(result, filename, local)
    return result

###############################################################################
#Export
###############################################################################
def export_code(code, filename='code1.py', local=False):
    global model_functions_directory
    filename1 = filename if local else join(model_functions_directory, filename)
    
    with open(filename1, 'w') as file:
        file.write(code)
    return
    
###############################################################################
#Supporting Functions
############################################################################### 
def states_to_code(states, indent=1):
    longest = len(max(states, key=len))
    temp    = ['\t'*indent + states[i] + ' '*(longest - len(states[i]) + 1) + '= y[' + str(i) + ']' for i in range(len(states))]
    result  = '\n'.join(temp)
    
    return result
    
def params_to_code(params, indent=1):
    longest = len(max(params, key=len))
    temp    = ['\t'*indent + params[i] + ' '*(longest - len(params[i]) + 1) + '= params[' + str(i) + ']' for i in range(len(params))]
    result  = '\n'.join(temp)
    
    return result

def equations_to_code(equations, states, indent=1):
    '''
    Excepts a list of strings corresponding to equations and formats them.
    Assumes the differentials are given by "d"+state_name.
    
    E.g. If the states are x1 and x2, the return value is np.array([x1, x2])
    
    '''
    result = ''
    blocks = []
    lhs    = []
    rhs    = []
    diff   = {'d'+state: False for state in states}
    for equation in equations:
        if equation.strip():
            lhs_, rhs_ = [s.strip() for s in equation.split('=')]
       
            lhs.append(lhs_)
            rhs.append(rhs_)
            
            if lhs_ in diff:
                diff[lhs_] = True
        else:
            if lhs:
                blocks.append([lhs.copy(), rhs.copy()])
                lhs = []
                rhs = []
            else:
                pass
    if lhs:
        blocks.append([lhs, rhs])
                
    temp = []
    for block in blocks:
        if temp:
            temp.append('\t'*indent)
        temp.append(equations_to_code_helper(*block, indent=indent))
    
    
    return_value = '\t'*indent + 'return np.array([' + ', '.join(diff.keys()) + '])'
    
    if not all(diff.values()):
        missing = [lhs_ for lhs_ in diff if not diff[lhs_]]
        message = 'The following differentials are missing: {missing}'.format(missing=missing)
        raise Exception(message)
        
    result  = '\n'.join(temp) + '\n\n' + return_value
    return result 

def equations_to_code_helper(diff, expr, indent=1):       
    longest = len(max(diff)) 
    temp    = ['\t'*indent + diff[i] + ' '*(longest - len(diff[i]) + 1) + '= ' + expr[i] for i in range(len(diff))]
    result  = '\n'.join(temp)
    return result
    
if __name__ == '__main__':
    __model__ = {'system_type' : 'Inducible, ConstInd',    
                 'variables'   : ['Ind', 'mRNA', 'Pep', 'syn_mRNA', 'deg_mRNA', 'syn_Pep', 'deg_Pep', 'Ki'],
                 'states'      : ['mRNA', 'Pep'], 
                 'parameters'  : ['syn_mRNA', 'deg_mRNA', 'syn_Pep', 'deg_Pep', 'Ki'],
                 'inputs'      : ['Ind'],
                 'equations'   : ['dmRNA  = syn_mRNA*Ind/(Ind + Ki)     - deg_mRNA*mRNA',
                                  'dPep   = syn_Pep*mRNA - deg_Pep',
                                  'dummy  = log(2.7)',
                                  'dPep   = dummy'
                                  ]
                 }
    
    equations = __model__['equations']
    states    = __model__['states']
    
    equations = equations[::-1]
    
    # print(equations_to_code(equations, states))
    code = model_to_code(__model__,  use_numba=True, local=False, mode='')
    print(code)
    exec(code)
    d = model_Inducible_ConstInd(np.array([1, 1]), 0, np.array([0.1, 0.1, 0.1, 0.1, 0.1]))
    assert all(np.isclose(d, [-0.1, 0.43136], atol=1e-3))


import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy             as np
import pandas            as pd
import seaborn           as sns
from SALib.analyze import sobol, fast, rbd_fast, delta
from SALib.sample import saltelli, fast_sampler, latin

###############################################################################
#Non-Standard Imports
###############################################################################
try:
    from . import simulation as sim
except:
    import simulation as sim
    
###############################################################################
#High Level Wrappers
###############################################################################
def analyze(params, models, fixed_parameters, objective, parameter_bounds={}, mode='np', analysis_type='sobol', N=256):
    '''Main algorithm for sensitivity analysis. Wraps analyze_sensitivty and sample_and_integrate.
    '''
    
    #Check for input errors
    if len(parameter_bounds) and len(parameter_bounds) != len(params.columns) - len(fixed_parameters):
        raise Exception('len(parameter_bounds) not equal to len(params) - len(fixed_parameters)')

    if not all(list(map(lambda x: x in params, fixed_parameters))):
        raise Exception('One or more unexpected parameters in fixed_parameters.' )
    
    if mode not in ['np' , 'df']:
        raise Exception('mode must be np or df')
    
    if analysis_type not in ['sobol', 'fast', 'rbd-fast' ,'delta']:
        raise Exception('analysis_type must be sobol, fast or delta.')
        
    #Generate samples and integrate
    em, samples, problems = sample_and_integrate(models, params, fixed_parameters, parameter_bounds, objective, analysis_type=analysis_type, N=N)
    
    #Analyze results
    analysis_result  = analyze_sensitivity(em, samples, problems, analysis_type=analysis_type)
    
    return analysis_result, em, samples, problems

def sample_and_integrate(models, params, fixed_parameters, parameter_bounds, objective, analysis_type='sobol', N=256):
    '''
    :meta private:
    '''
    samples, problems = make_samples(models, params, fixed_parameters, parameter_bounds, analysis_type=analysis_type, N=N)
    em                = integrate_samples(models, samples, objective)
    
    return em, samples, problems

###############################################################################
#Output Evaluation and Sensitivity Analysis
###############################################################################
def analyze_sensitivity(objective_function_values, samples, problems, analysis_type='sobol', **kwargs):
    '''
    :meta private:
    '''
    s = {}
    e = objective_function_values
    
    for model_num in e:
        s[model_num] = {}  
        for scenario in e[model_num]:
            s[model_num][scenario] = {}
            for row_name in e[model_num][scenario]:
                s[model_num][scenario][row_name] = {}
                for func in e[model_num][scenario][row_name]:
                    problem  = problems[model_num][row_name]
                    permuted = problem['names'] 
                    x        = samples[model_num][row_name][permuted]
                    y_val    = e[model_num][scenario][row_name][func]
                    
                    s[model_num][scenario][row_name][func] = salib_wrapper(problem, y_val, x, analysis_type, **kwargs)
    
    return s

def salib_wrapper(problem, y_val, x, analysis_type='sobol', **kwargs):
    '''
    Backend wrapper for sobol, fast and delta analysis.
    
    :meta private:
    '''
    if analysis_type == 'sobol':
        return sobol.analyze(problem, y_val, **kwargs)
    elif analysis_type == 'fast':
        return fast.analyze(problem, y_val, **kwargs)
    elif analysis_type == 'delta':
        return delta.analyze(problem, x, y_val, **kwargs)
    elif analysis_type == 'rbd-fast':
        return rbd_fast.analyze(problem, x, y_val, **kwargs)
    else:
        raise Exception('Could not find analyzer. analysis_type must be sobol, fast, rbd-fast or delta.')    
   

###############################################################################
#Sampling and Integration 
###############################################################################
def integrate_samples(models, samples, objective, args=()):
    '''
    Integrates the samples and evaluates the objective function for each.

    Parameters
    ----------
    models : dict
        A dictionary of model data structures.
    samples : dict
        A dictionary of parameter values for sampling.
    objective : dict
        A dictionary of objective functions to evaluate indexed by model_num.
    args : tuple, optional
        Additional arguments for integration. The default is ().

    Returns
    -------
    e_models : dict
        The objective function values.
    '''
    e_models = {}
    samples1 = {}
    for model_num in samples:
        samples1[model_num] = {}
        for row_name in samples[model_num]:
            if type(samples[model_num][row_name]) != pd.DataFrame:
                try:
                    samples1[model_num][row_name] = pd.DataFrame(samples[model_num][row_name])
                except:
                    samples1[model_num][row_name] = pd.DataFrame([samples[model_num][row_name]])
            else:
                samples1[model_num][row_name] = samples[model_num][row_name]
                
    for model_num in models:
        model               = models[model_num]
        e_models[model_num] = {}
        for scenario_num in models[model_num]['init']:
            if type(scenario_num) == int:
                if scenario_num < 1:
                    continue
            e_models[model_num][scenario_num] = {}
            for row_name in samples[model_num]:
                params1 = samples[model_num][row_name]
                temp    = {func: [] for func in objective.get(model_num)}
                
                e_models[model_num][scenario_num][row_name] = {}      
                for _, row in params1.iterrows():
                    y_model, t_model = sim.piecewise_integrate(params        = row.values,
                                                               function      = model['function'], 
                                                               init          = model['init'][scenario_num],
                                                               tspan         = model['tspan'],
                                                               modify_init   = model['int_args']['modify_init'],
                                                               modify_params = model['int_args']['modify_params'],
                                                               solver_args   = model['int_args']['solver_args'],
                                                               model_num     = model_num, 
                                                               scenario_num  = scenario_num, 
                                                               overlap       = True,
                                                               *args
                                                               )
                   
                
                    for func in objective.get(model_num):
                        y_model_ = pd.DataFrame(y_model) if 'mode=pd' in str(func.__doc__) else y_model
                        variable = func(y_model_, t_model, row.values)
                        temp[func].append(variable)
                
                for func in objective.get(model_num):
                    e_models[model_num][scenario_num][row_name][func] = np.array(temp[func])
    
        return e_models
    
###############################################################################
#Sample Generation
###############################################################################
def make_samples(models, params, fixed_parameters, parameter_bounds, analysis_type='sobol', N=256):
    '''Generates samples for each model.
    '''
    try:
        params1 = pd.DataFrame(params)
    except:
        params1 = pd.DataFrame([params])
    
    samples  = {}
    problems = {}
    
    for model_num in models:
        samples[model_num]  = {}
        problems[model_num] = {}
        
        for name, row in params1.iterrows():
            row_samples, row_problem = make_samples_for_row(row, 
                                                            models[model_num], 
                                                            fixed_parameters, 
                                                            parameter_bounds, 
                                                            analysis_type=analysis_type, 
                                                            N=N
                                                            )
            samples[model_num][name]  = row_samples
            problems[model_num][name] = row_problem
    
    return samples, problems

def make_samples_for_row(row, model, fixed_parameters, parameter_bounds={}, analysis_type='sobol', N=256, **kwargs):
    '''
    Supporting function for sample_and_integrate. Do not run.
    row must be a pandas Series
    
    :meta private:
    '''

    param_names      = model['params']
    base_value       = row[param_names]
    to_permute       = [p for p in param_names if p not in fixed_parameters]
    to_keep          = [p for p in param_names if p     in fixed_parameters]
    if parameter_bounds:
        bounds = np.array([parameter_bounds[p] for p in to_permute])
    else:
        bounds = [[base_value[p]/10, base_value[p]*10] for p in to_permute] 
    
    problem = {'num_vars' : len(to_permute),
               'names'    : to_permute,
               'bounds'   : bounds,
               }
    
    if analysis_type == 'sobol':          
        new_values   = saltelli.sample(problem, N, **kwargs)
    elif analysis_type == 'fast':   
        new_values   = fast_sampler.sample(problem, N, **kwargs)
    elif analysis_type == 'delta':
        new_values   = latin.sample(problem, N)
    elif analysis_type == 'rbd-fast':
        new_values   = latin.sample(problem, N)
    else:
        raise Exception('Could not find analyzer. analysis_type must be sobol, fast, rbd-fast or delta.')    
        
    new_values   = pd.DataFrame(new_values, columns=to_permute)
    
    fixed_parametersped_values = np.repeat(base_value[to_keep].values[np.newaxis,:], len(new_values), 0)
            
    samples = np.concatenate((new_values, fixed_parametersped_values), axis=1)
    samples = pd.DataFrame(samples, columns = to_permute+to_keep)
    samples = samples[param_names]
    
    return samples, problem

###############################################################################
#Visualization
###############################################################################
def plot_first_order(analysis_result, problems={}, titles={}, analysis_type='sobol', figs=None, AX=None, analysis_keys=(), **heatmap_args):
    '''
    If analysis_type is neither sobol, fast nor delta, use analysis_keys to specify which 
    keys/columns to use.
    '''
    
    #Fix this
    figs1, AX1 = (figs, AX) if AX else make_AX1(analysis_result) 

    if analysis_type == 'sobol':
        keys = ['S1', 'S1_conf', 'ST', 'ST_conf']
    elif analysis_type == 'fast':
        keys = ['S1', 'ST']
    elif analysis_type == 'delta':
        keys = ['delta', 'delta_conf', 'S1', 'S1_conf']
    elif analysis_type == 'rbd-fast':
        keys = ['S1']
    else:
        if analysis_keys:
            keys = analysis_keys
        else:
            raise Exception('analysis_key must be specified if using custom analysis.')
    
    for model_num in analysis_result:
        for scenario_num in analysis_result[model_num]:
            for row_name in analysis_result[model_num][scenario_num]:
                for func in analysis_result[model_num][scenario_num][row_name]:
                    
                    r  = analysis_result[model_num][scenario_num][row_name][func]
                    p  = problems[model_num][row_name]['names'] if problems else None
                    df = pd.DataFrame({key: r[key] for key in keys}, columns=keys, index=p)
                    ax = AX1[model_num][scenario_num][row_name][func]
                    
                    sns.heatmap(df, ax=ax, **heatmap_args)
                    
                    try:
                        title = titles[model_num][scenario_num][row_name][func]
                    except:
                        try:
                            title = titles[model_num][scenario_num][func]
                        except:
                            if titles.get(model_num, {}).get(func, None):
                                title = make_default_title(model_num, scenario_num, row_name, titles[model_num][func])
                            else:
                                title = make_default_title(model_num, scenario_num, row_name, func.__name__)
                    ax.set_title(title)
    
    [sim.fs(fig) for fig in figs1]
    
    return figs1, AX1
    

def plot_second_order(analysis_result, problems={}, titles={}, analysis_type='sobol', figs=None, AX=None, analysis_keys=(), **heatmap_args):
    '''
    If analysis_type is not sobol, use analysis_keys to specify which keys/columns to use.
    '''

    if analysis_type == 'sobol':
        keys = ['S2', 'S2_conf']
    else:
        if analysis_keys:
            keys = analysis_keys
        else:
            raise Exception('analysis_key must be specified if using custom analysis.')
    
    figs1, AX1 = (figs, AX) if AX else make_AX2(analysis_result, keys) 

    for model_num in analysis_result:
        for scenario_num in analysis_result[model_num]:
            for row_name in analysis_result[model_num][scenario_num]:
                for func in analysis_result[model_num][scenario_num][row_name]:
                    r  = analysis_result[model_num][scenario_num][row_name][func]
                    p  = problems[model_num][row_name]['names'] if problems else None
                    
                    for key in keys:
                        ax = AX1[model_num][scenario_num][row_name][func][key]
                        df = pd.DataFrame(r[key], columns=p, index=p)
                    
                        sns.heatmap(df, ax=ax, **heatmap_args)
                    
                        try:
                            title = titles[model_num][scenario_num][row_name][func]
                        except:
                            try:
                                title = titles[model_num][scenario_num][func]
                            except:
                                if titles.get(model_num, {}).get(func, None):
                                    title = make_default_title(model_num, scenario_num, row_name, titles[model_num][func])
                                else:
                                    title = make_default_title(model_num, scenario_num, row_name, func.__name__)
                        ax.set_title(title + ' (' + key + ')')
    
    [sim.fs(fig) for fig in figs1]
    
    return figs1, AX1

def make_AX1(analysis_result):
    '''
    :meta private:
    '''
    figs = []
    AX   = {}
    for model_num in analysis_result:
        AX[model_num] = {}
        for scenario_num in analysis_result[model_num]:
            AX[model_num][scenario_num] = {}
            for row_name in analysis_result[model_num][scenario_num]:
                AX[model_num][scenario_num][row_name] = {}
                
                n    = len(analysis_result[model_num][scenario_num][row_name])
                rows = int(n**0.5)
                cols = int(n/rows + 0.5)
                fig  = plt.figure()
                AX_  = [fig.add_subplot(rows, cols, i+1) for i in range(n)] 
                i    = 0
                figs.append(fig)
                for func in analysis_result[model_num][scenario_num][row_name]:
                    AX[model_num][scenario_num][row_name][func] = AX_[i]
                    i += 1
    return figs, AX

def make_AX2(analysis_result, keys):
    '''
    :meta private:
    '''
    figs = []
    AX   = {}
    for model_num in analysis_result:
        AX[model_num] = {}
        for scenario_num in analysis_result[model_num]:
            AX[model_num][scenario_num] = {}
            for row_name in analysis_result[model_num][scenario_num]:
                AX[model_num][scenario_num][row_name] = {}

                for func in analysis_result[model_num][scenario_num][row_name]:
                    fig  = plt.figure()
                    AX_  = [fig.add_subplot(1, 2, i+1) for i in range(2)] 
                    AX[model_num][scenario_num][row_name][func] = dict(zip(keys, AX_))
                    figs.append(fig)

    return figs, AX

def make_default_title(model_num, scenario_num, row_name, func_name):
    '''
    :meta private:
    '''
    return 'Model {model_num}, Scenario {scenario_num}, Row {row_name}, {func}'.format(model_num    = model_num, 
                                                                                  scenario_num = scenario_num,
                                                                                  row_name     = row_name,
                                                                                  func         = func_name
                                                                                  )



from .OnlinetoConfig import (get_online_biomodel, 
                             gen_settingstemplate,
                             gen_config,
                            )
from .sbmlgen        import (database_to_sbml, 
                             autogenerate_sbml_from_folder, 
                            )
from .combinegen     import *
from .               import OnlinetoConfig
from .               import sbmlgen 

sbml_to_config = lambda *args, **kwargs: OnlinetoConfig.sbmltoconfig(*args, **kwargs)[0]

def sbml_to_config(sbmlfile, system_type, output_path='', tspan=None, is_path=True):
    '''Converts SBML file to config file(s).
    
    :param sbmlfile: Path-like or SBML string.
    :param system_type: The system_type for naming the model.
    :param output_paths: The location for creating the config files. No files will be created if the string is empty.
    :param tspan: A list/array of time points.
    :param is_path: True if sbmlfile is a path, False if sbmlfile is an SBML string.
    :return result: Config string.
    
    '''
    if is_path:
        
        with open(sbmlfile, 'r') as file:
            sbmlstr = file.read()
    else:
        sbmlstr = sbmlfile
        
    result  = OnlinetoConfig.sbmltoconfig(sbmlstr, system_type, output_path, tspan)
    return result

def config_to_sbml(inifile, output_path=''):
    '''Converts config file(s) to SBML.
    
    :param inifile: Path-like or ini string.
    :param output_paths: The location for creating the config files. No files will be created if the string is empty.
    :return result: List of output files.
    
    '''
    return sbmlgen.config_to_sbml([inifile], output_path)


# -*- coding: utf-8 -*-
"""
This module contains functions to generate SBML standard files.

Features:
-to generate SBML .xml files from a list of configuration files
-to generate SBML files from model database
-to autogenerate SBML .xml files for all configuration files contained inside a folder path  
"""

import os
import glob

import BMSS.sbml.simplesbml as simplesbml
import BMSS.models.model_handler    as mh
import BMSS.models.settings_handler as sh

def unitlookup(settings):
    '''Reads in the list of units listed in the database and converts them to 
    units defined in simplesbml
    
    :param settings:  settings database of model
    :return: cleaned units of parameters in list format
    
    :meta private:
    '''
    
    unit_model = settings['units']
    
    #--Checker--
    unit_model_set = set(unit_model.keys())
    params_set     = set(settings['parameters'].keys())
    
    missing    = params_set.difference(unit_model_set)
    unexpected = unit_model_set.difference(params_set)
    
    if missing:
        raise Exception(f'No units provided for: {missing}')
    elif unexpected:
        raise Exception(f'Unexpected units provided for: {unexpected}')
    
    for w in unit_model: #Unit Look-up/conversion
        unit_model[w] = unit_model[w].replace('aa', 'items') 
        unit_model[w] = unit_model[w].replace('mol-1', 'permol')
        unit_model[w] = unit_model[w].replace('molmin-1', 'molpermin')
        unit_model[w] = unit_model[w].replace('1/', 'per_') 
        unit_model[w] = unit_model[w].replace('M/', 'molper') 
        unit_model[w] = unit_model[w].replace('NONE', 'Dimension_less') 
        unit_model[w] = unit_model[w].replace('dimensionless', 'Dimension_less')
        unit_model[w] = unit_model[w].replace('/l', 'L') 
        unit_model[w] = unit_model[w].replace('sec', 'second')
        unit_model[w] = unit_model[w].replace('molL-1min-1', 'molperLmin') 
        unit_model[w] = unit_model[w].replace('min-1', 'per_min')
        unit_model[w] = unit_model[w].replace('molL-1', 'molperL')   
        
    return unit_model

def SBMLcreation(core_model, settings, unit_model, addparam, init_scenario, param_scenario):
    '''Reads in the core model, settings, parameters and which scenario of 
    init and parameters values and outputs SBML format in sbmlstr
    
    :param core_model:  core model of model in nested dictionary format
    :param settings:  settings database of model
    :param unit_model:  units for parameters in list format
    :param addparam:  parameters of model in dataframe
    :param init_scenario:  init value scenario in int format
    :param param_scenario:  parameter value scenario in int format
    :return: SBML string in XML format
    :meta private:
    '''
    
    #--Checker--
    if core_model['system_type'] != settings['system_type']:
        raise Exception('System_type not the same between the core_model and settings')
        
    if 'inputs' in core_model.keys():
        if (len(core_model['parameters'])+len(core_model['inputs'])) != len(addparam.columns):
            raise Exception('settings have missing parameters')
    else:
        if len(core_model['parameters']) != len(addparam.columns):
            raise Exception('settings have missing parameters')
    
        
    
    
    model_sbml = simplesbml.SbmlModel()      
    
    for i, state in enumerate(core_model['states']):
        amt = float(settings['init'].loc[init_scenario][i])
        model_sbml.addSpecies(state, amt)    
    
    for i, name in enumerate(addparam.columns):
        val   = addparam.values[param_scenario,i]
        units = unit_model[name]
        model_sbml.addParameter(name, val, units = units)
    
    for eqn in core_model['equations']:
        if eqn:
            if eqn[0] == 'd': #if equation is for deritive with time
                variable_start  = eqn.index("d")
                variable_end    = eqn.index("=")
                eqn = eqn.replace('**', '^')  
                variable_name   = eqn[variable_start+1:variable_end]
                variable_name   = variable_name.replace(' ', '')
                model_sbml.addRateRule(variable_name, eqn[variable_end+1:])
                
            else: #if equation is for variable directly
                variable_end    = eqn.index("=")
                assignment_name = eqn[:variable_end]
                assignment_name = assignment_name.replace(' ', '')
                eqn             = eqn.replace('**', '^') 
                model_sbml.addAssignmentRule(assignment_name, eqn[variable_end+1:])
                
    sbmlstr = model_sbml.toSBML()
    return sbmlstr

  
def config_to_sbml(inifileslist, output_path=''):
    '''Read in list of configuration files and generate the corresponding SBML
    files.
    
    :param inifileslist: a list of input configuration files in strings.
    :type inifileslist: list
    '''
    
    sbmlfilelist=[]
    number_scenario = 0
    
    for f in inifileslist:
        print('\n', f)
    
       
    for f in inifileslist:    #Add ini files to database
        print('filename', f)
        system_type                 = mh.config_to_database(f)
        system_types_settings_names = sh.config_to_database(f)
        system_type, settings_name  = system_types_settings_names[0]
        search_result_model         = mh.quick_search(system_type)
        search_result_settings      = sh.quick_search(system_type=system_type, settings_name=settings_name)
        core_model                  = search_result_model
        settings                    = search_result_settings
        addparam                    = settings['parameters']
        number_init                 = settings['init'].index
        number_parameters           = len(settings['parameters'])
        unit_model                  = unitlookup(settings)
    
        for j in number_init: #Cycle through number of init values
            for k in range(number_parameters): #Cycle through number of parameters
                sbmlstr             = SBMLcreation(core_model, settings, unit_model, addparam, j, k)
                #Function that creates SBML file and returns the number of files outputed and the file list.
                number_scenario     = number_scenario + 1
                placeholder         = os.path.splitext(f)[0]
                sbmlfilename        = os.path.join(output_path, 'DatabasetoSBML_' + placeholder +'.xml')
                f                   = open(sbmlfilename, 'w') #creates SBML file in same folder as python script
                f.write(sbmlstr)
                f.close()
                del sbmlstr
                sbmlfilelist.append(sbmlfilename)
    
    print()
    print('Number of Files Printed =', number_scenario)
    print('SBML files outputed: ', sbmlfilelist)

    return sbmlfilelist

def autogenerate_sbml_from_folder(folderpath, output_path=''):
    '''To automatically generate the corresponding SBML files from the given
    folder containing all the configuration files.
    
    :param folderpath: the path to the folder containing all the configuration files.
    :type folderpath: str
    '''
    if os.path.exists(folderpath) != True:
        raise Exception("Input folder does not exist") 
    
    #-- Makes output folder if not already created -- 
    try:
        os.mkdir(output_path)
    except OSError:
        print('Output Folder Exists')
    else:
        print('Folder created: \n', output_path)
        
    files = [f for f in glob.glob(os.path.join(folderpath,"**/*.ini"), recursive=True)]
    
    for f in files:
        print('\n', f)

    sbmlfilelist=[]
    number_scenario = 0
       
    for f in files:    #Add ini files to database
        system_type                 = mh.config_to_database(f)
        system_types_settings_names = sh.config_to_database(f)
        system_type, settings_name  = system_types_settings_names[0]
        search_result_model         = mh.quick_search(system_type)
        search_result_settings      = sh.quick_search(system_type=system_type, settings_name=settings_name)
        core_model                  = search_result_model
        settings                    = search_result_settings
        addparam                    = settings['parameters']
        number_init                 = settings['init'].index
        number_parameters           = len(settings['parameters'])
        unit_model                  = unitlookup(settings)
       
        for j in number_init: #Cycle through number of init values
            for k in range(number_parameters): #Cycle through number of parameters   
                sbmlstr             = SBMLcreation(core_model, settings, unit_model, addparam, j, k)
                number_scenario     = number_scenario + 1
                placeholder         = os.path.splitext(os.path.basename(f))[0]
                sbmlfilename        = os.path.join(output_path, 'DatabasetoSBML_' + placeholder +'.xml')
                f                   = open(sbmlfilename, 'w') #creates SBML file in same folder as python script
                f.write(sbmlstr)
                f.close()
                del sbmlstr
                sbmlfilelist.append(sbmlfilename)
        
        #Function that creates SBML file and returns the number of files outputed and the file list.
        
    
    print()
    print('Number of Files Printed =', number_scenario)
    print('SBML files outputed: ', sbmlfilelist)
    
    return sbmlfilelist

def database_to_sbml(system_type, settings_name, output_path=''):
    '''Select model stored in database and generate the SBML file. 
    
    :param system_type:  system_type of model to search in database in string format
    :param settings_name:  settings_name of model to search in database in string format
    '''
    sbmlfilelist            = []
    number_scenario         = 0

    
    search_result_settings  = sh.search_database(system_type, settings_name)
    search_result_model     = mh.quick_search(system_type)
    core_model              = search_result_model
    settings                = search_result_settings[0]
    addparam                = settings['parameters']
    number_init             = settings['init'].index
    number_parameters       = len(settings['parameters'])
    
    unit_model = unitlookup(settings)
    
    for j in number_init: #Cycle through number of init
        for k in range(number_parameters): #Cycle through number of parameters
            sbmlstr         = SBMLcreation(core_model, settings, unit_model, addparam, j, k)           
            number_scenario = number_scenario + 1
            sbmlfilename    = os.path.join(output_path, 'DatabasetoSBML_' + str(number_scenario)+'.xml')
            f               = open(sbmlfilename, 'w')
            f.write(sbmlstr)
            f.close()
            del sbmlstr
            sbmlfilelist.append(sbmlfilename)
    #Function that creates SBML file and returns the number of files outputed and the file list.
        
    print()
    print('Number of Scenarios in Model =', number_scenario)
    print('SBML files outputed: ', sbmlfilelist)
    
    return sbmlfilelist